/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_spi
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     33
// Number of 4 input LUTs:         54
// Number of occupied Slices:      42
// Total Number of 4 input LUTs:   62
// Number of bonded IOBs :         29
//   IOB Flip Flops :               -
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :           -          -

module rtavr_ior_spi(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, input SS_DDR
		, input SS_IN
		, input SCK_IN
		, input MOSI_IN
		, input MISO_IN
		, output SPI_MASTER
		, output SPI_SLAVE
		, output SCK_OUT
		, output MOSI_OUT
		, output MISO_OUT
		// interrupt
		, output SPI_INT
		, input RST_SPI_INT
		);
  parameter BASE_ADDR  = 6'h2e;    // SPI : SPDR 0x2e, SPSR, SPCR

  reg  [6:0] r_prescaler;
  reg  [2:0] r_count;
  reg  snd_data_rdy;
  reg  r_count_en;
  reg  r_count_en_prev;
  reg  r_ss;
  reg  r_ss_prev;
  reg  r_spsr_lck;
  reg  r_spi_mclk;
  reg  r_spi_in;
  reg  r_int_in;
  reg  r_int_prev;
  
  reg  [7:0] r_spdr;
  reg  [7:0] r_spcr;
  reg  r_spif; // spsr bit7
  reg  r_wcol; // spsr bit6
  reg  r_spi2x; // spsr bit0

  wire SPIE = r_spcr[7];
  wire SPE = r_spcr[6];
  wire DORD = r_spcr[5];
  wire MSTR = r_spcr[4];
  wire CPOL = r_spcr[3];
  wire CPHA = r_spcr[2];
  wire [2:0] SPR = { r_spcr[1:0] , ~r_spi2x };

  wire OE_spdr = RD & (ADDR == BASE_ADDR);
  wire WE_spdr = WR & (ADDR == BASE_ADDR);
  wire OE_spsr = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_spsr = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_spcr = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_spcr = WR & (ADDR == (BASE_ADDR + 2));

  wire [7:0] i_DO = OE_spdr ? r_spdr[7:0] 
		: OE_spsr ? { r_spif, r_wcol, 5'b0, r_spi2x }
		: OE_spcr ? r_spcr: 0;
`ifdef OUTPUT_MUX_TRI
  bufif1(DO, i_DO, (OE_spdr | OE_spsr | OE_spcr));
`else
  assign DO = i_DO;
`endif

  wire v_spi_mclk = (SPR == 0) ? r_prescaler[0]
  		  : (SPR == 1) ? r_prescaler[1]
  		  : (SPR == 2) ? r_prescaler[2]
  		  : (SPR == 3) ? r_prescaler[3]
  		  : (SPR == 4) ? r_prescaler[4]
  		  : (SPR == 5) | (SPR == 6) ? r_prescaler[5]
  		  :r_prescaler[6];

  wire v_spi_out = DORD ? r_spdr[0] : r_spdr[7] ;

  assign SPI_MASTER = SPE & MSTR;
  assign SPI_SLAVE = SPE & ~MSTR;
  assign MOSI_OUT = v_spi_out;
  assign MISO_OUT = v_spi_out;
  assign SPI_INT = r_spif & SPIE;

  wire v_spi_in = MSTR ? MISO_IN : MOSI_IN;

  wire v_spi_clk = MSTR ? v_spi_mclk : (SCK_IN ^ CPOL);
  reg  r_spi_clk;

  wire v_count_en = SPE & ~SS_IN & (~MSTR | MSTR & snd_data_rdy);

  wire do_count = v_count_en & (r_spi_clk ^ v_spi_clk);
  wire do_in  =  do_count & r_spi_clk;
  wire do_sft =  do_count & ~r_spi_clk;

  assign SCK_OUT = ( v_spi_mclk ^ CPOL ) & v_count_en;

  wire cond_spif_1 = ~WE_spdr & r_count_en & do_sft & (r_count == 7);
  wire cond_spif_2 = ~r_ss_prev & r_ss;
  wire cond_wcol   =  WE_spdr & r_count_en;

  always @(posedge CLK)
  begin
    if (~RESET | ~SPE)
      begin
	r_prescaler	<= 0;
	r_count_en	<= 0;
	r_count_en_prev <= 0;
	r_ss		<= 0;
	r_ss_prev	<= 0;
      end
    else
      begin
	r_prescaler	<= r_prescaler + 1;
	r_count_en	<= v_count_en;
	r_count_en_prev <= r_count_en;
	r_ss		<= SPE & MSTR & ~SS_DDR & ~SS_IN;
	r_ss_prev	<= r_ss;
      end
  end

  always @(posedge CLK)
  begin

    if (RESET == 0)
      begin
	r_spdr <= 0;
	r_spcr <= 0;
	r_spif <= 0;
	r_wcol <= 0;
	r_spi2x <= 0;
  	r_spi_clk <= 0;
	r_count <= 0;
	r_spi_in <= 0;
	r_spi_mclk <= 0;
	r_spsr_lck <= 0;
	snd_data_rdy <= 0;
      end
    else
      begin
  	r_spi_clk <= v_spi_clk;
	
	if (r_count_en & ~r_count_en_prev) r_count <= 0;

	if (WE_spdr)
	  begin
	    r_spdr <= DI;
	    snd_data_rdy <= 1;
	  end
	else 
	  begin
	    if (do_sft)
	      begin
		r_spi_mclk <= ~CPOL;
		r_count <= r_count + 3'b001;
		r_spdr <= DORD ? { r_spi_in , r_spdr[7:1]}
			       : { r_spdr[6:0], r_spi_in};
		if (r_count == 7) snd_data_rdy <= 0;
	      end
	    else if (do_in)
	      begin
		r_spi_mclk <= CPOL;
		r_spi_in <= v_spi_in;
	      end
	  end

	if(OE_spsr & (r_spif | r_wcol)) r_spsr_lck <= 0;
	else if (cond_spif_1 | cond_spif_2 | cond_wcol) r_spsr_lck <= 1;

	if(RST_SPI_INT | ~r_spsr_lck & (WE_spdr | OE_spdr) ) r_spif <= 0;
	else if (cond_spif_1 | cond_spif_2) r_spif <= 1;

	if(~r_spsr_lck & (WE_spdr | OE_spdr) ) r_wcol <= 0;
	else if (cond_wcol) r_wcol <= 1;

	if(WE_spcr) r_spcr <= DI;
//	if(WE_spsr) r_spif <= DI[7];
//	if(WE_spsr) r_wcol <= DI[6];
	if(WE_spsr) r_spi2x <= DI[0];
      end
  end

endmodule
